# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Composite filter implementations."""
from __future__ import annotations

import collections.abc as cabc
import functools
import typing as t

from capellambse import diagram

from .. import _semantic
from . import phase2_composite_filter


def _lookup_styleclasses(*styleclasses: str) -> frozenset[str]:
    returnclasses = set()
    for styleclass in styleclasses:
        new_styleclass, _ = _semantic.STYLECLASS_LOOKUP[styleclass]
        if new_styleclass:
            returnclasses.add(styleclass)
    return frozenset(returnclasses)


PORT_CL_COMPONENT = _lookup_styleclasses("ComponentPort")
PORT_CL_FUNCTION = _lookup_styleclasses(
    "FunctionInputPort", "FunctionOutputPort"
)


def hide_empty_ports(
    ebd: t.Any | None,
    dgobject: diagram.DiagramElement,
    *,
    classes: cabc.Container[str],
) -> None:
    """Hide child ports without context.

    This filter utilizes the fact that boxes track all connected edges
    as their "context". As ports do not have children, their context can
    only be the edge(s) that connect directly to them.
    """
    del ebd
    assert isinstance(dgobject, diagram.Box)

    for child in dgobject.children:
        if child.port and child.styleclass in classes and not child.context:
            child.hidden = True


phase2_composite_filter("Hide Component Ports without Exchanges")(
    functools.partial(hide_empty_ports, classes=PORT_CL_COMPONENT)
)
phase2_composite_filter("Hide Function Ports without Exchanges")(
    functools.partial(hide_empty_ports, classes=PORT_CL_FUNCTION)
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=abstract-method, useless-suppression
# For some reason, pylint in Github CI didn't get the memo that these aren't
# actually abstract methods. Other pylint installations seem to agree that
# implementing these methods isn't necessary. So we just ignore the warning
# about that here.
# TODO Revisit this decision some time in the future

from __future__ import annotations

import collections.abc as cabc
import itertools
import os
import pathlib
import re
import typing as t
import urllib.parse

import requests

from capellambse import helpers, loader

from . import abc


class DownloadStream(t.BinaryIO):
    __stream: cabc.Iterator[bytes]
    __buffer: memoryview

    def __init__(
        self, session: requests.Session, url: str, chunk_size: int = 1024**2
    ) -> None:
        self.url = url
        self.chunk_size = chunk_size

        response = session.get(self.url, stream=True)
        if response.status_code == 404:
            raise FileNotFoundError(url)
        response.raise_for_status()
        self.__stream = response.iter_content(
            self.chunk_size, decode_unicode=False
        )
        self.__buffer = memoryview(b"")

    def __enter__(self) -> DownloadStream:
        return self

    def __exit__(self, *args: t.Any) -> None:
        self.close()

    def read(self, n: int = -1) -> bytes:
        if n == -1:
            return b"".join(itertools.chain((self.__buffer,), self.__stream))

        if not self.__buffer:
            try:
                self.__buffer = memoryview(next(self.__stream))
            except StopIteration:
                return b""

        chunk = bytes(self.__buffer[:n])
        self.__buffer = self.__buffer[n:]
        return chunk

    def readable(self) -> bool:
        return True

    def write(self, s: bytes | bytearray) -> int:  # type: ignore[override]
        raise TypeError("Cannot write to a read-only stream")

    def writable(self) -> bool:
        return False

    def close(self) -> None:
        del self.__stream
        del self.__buffer


class HTTPFileHandler(abc.FileHandler):
    """A remote file handler that fetches files using HTTP GET."""

    def __init__(
        self,
        path: str | os.PathLike,
        username: str | None = None,
        password: str | None = None,
        *,
        headers: (
            dict[str, str]
            | requests.structures.CaseInsensitiveDict[str]
            | None
        ) = None,
        subdir: str | pathlib.PurePosixPath = "/",
    ) -> None:
        """Connect to a remote server through HTTP or HTTPS.

        This file handler supports three ways of specifying a URL:

        1. If a plain URL is passed, the requested file name is appended
           after a forward slash ``/``.
        2. If the URL contains ``%s``, it will be replaced by the
           requested file name, instead of appending it at the end. This
           allows for example to pass query parameters after the file
           name. File names are percent-escaped as implemented by
           ``urllib.parse.quote``.
        3. The sequence ``%q`` is replaced similar to ``%s``, but the
           forward slash ``/`` is not considered a safe character and is
           percent-escaped as well.

        Examples: When requesting the file name ``demo/my model.aird``,
        ...

        - ``https://example.com/~user`` as ``path`` results in the URL
          ``https://example.com/~user/demo/my%20model.aird``
        - ``https://example.com/~user/%s`` results in
          ``https://example.com/~user/demo/my%20model.aird``
        - ``https://example.com/?file=%q`` results in
          ``https://example.com/?file=demo%2Fmy%20model.aird``

        Note that the file name that is inserted into the URL will never
        start with a forward slash. This means that a URL like
        ``https://example.com%s`` will not work; you need to hard-code
        the slash at the appropriate place.

        This also applies to the ``%q`` escape. If the server expects
        the file name argument to start with a slash, hard-code a
        percent-escaped slash in the URL. For example, instead of
        ``...?file=%q`` use ``...?file=%2F%q``.

        Parameters
        ----------
        path
            The base URL to fetch files from. Must start with
            ``http://`` or ``https://``. See above for how to specify
            complex URLs.
        username
            The username for HTTP Basic Auth.
        password
            The password for HTTP Basic Auth.
        headers
            Additional HTTP headers to send to the server.
        subdir
            Prepend this path to all requested files. It is subject to
            the same file name escaping rules explained above.
        """
        if not isinstance(path, str):
            raise TypeError(
                "HTTPFileHandler requires a str path, not"
                f" {type(path).__name__}"
            )
        if bool(username) != bool(password):
            raise ValueError(
                "Either both username and password must be given, or neither"
            )

        if "%s" not in path and "%q" not in path:
            path = path.rstrip("/") + "/%s"

        super().__init__(path, subdir=subdir)

        self.session = requests.Session()
        self.session.headers.update(headers or {})
        if username and password:
            self.session.auth = (username, password)

    def get_model_info(self) -> loader.ModelInfo:
        assert isinstance(self.path, str)
        parts = urllib.parse.urlparse(self.path)
        return loader.ModelInfo(
            title=parts.path.rsplit("/", maxsplit=1)[-1],
            url=self.path,
        )

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        if "w" in mode:
            raise NotImplementedError("Cannot upload to HTTP(S) locations")
        assert isinstance(self.path, str)
        fname = self.subdir / helpers.normalize_pure_path(filename)
        fname_str = str(fname).lstrip("/")
        replace = {
            "%s": urllib.parse.quote(fname_str, safe="/"),
            "%q": urllib.parse.quote(fname_str, safe=""),
        }
        url = re.sub("%[sq]", lambda m: replace[m.group(0)], self.path)
        assert url != self.path
        return DownloadStream(  # type: ignore[abstract] # false-positive
            self.session, url
        )

    def write_transaction(self, **kw: t.Any) -> t.NoReturn:
        raise NotImplementedError(
            "Write transactions for HTTP(S) are not implemented"
        )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=redefined-outer-name
from __future__ import annotations

import base64
import contextlib
import errno
import logging
import pathlib
import re
import shutil
import subprocess
import sys
from importlib import metadata
from unittest import mock

import pytest
import requests_mock

import capellambse
from capellambse.filehandler import gitlab_artifacts

# pylint: disable-next=relative-beyond-top-level, useless-suppression
from .conftest import TEST_MODEL, TEST_ROOT  # type: ignore[import]

TEST_MODEL_5_0 = TEST_ROOT / "5_0" / TEST_MODEL

DUMMY_SVG = b'<svg xmlns="http://www.w3.org/2000/svg"/>'
DUMMY_PNG = base64.standard_b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAACklEQVR4nGMAAQAABQABDQott"
    "AAAAABJRU5ErkJggg=="
)


@pytest.mark.parametrize(
    "path",
    [
        pytest.param(str(TEST_MODEL_5_0), id="From string"),
        pytest.param(TEST_MODEL_5_0, id="From path"),
    ],
)
def test_model_loading_via_LocalFileHandler(path: str | pathlib.Path):
    capellambse.MelodyModel(path)


@pytest.mark.parametrize("suffix", [".afm", ".capella"])
def test_model_loading_with_invalid_entrypoint_fails(suffix: str):
    with pytest.raises(ValueError, match="(?i)invalid entrypoint"):
        capellambse.MelodyModel(TEST_MODEL_5_0.with_suffix(suffix))


def test_model_loading_via_GitFileHandler():
    path = "git+" + pathlib.Path.cwd().as_uri()
    capellambse.MelodyModel(
        path, entrypoint="tests/data/melodymodel/5_0/Melody Model Test.aird"
    )
    assert not pathlib.Path.cwd().joinpath("capellambse.lock").exists()


def test_GitFileHandler_locks_repo_during_tasks(monkeypatch, caplog):
    did_ls_files = False

    def mock_run(cmd, *args, encoding="", **kw):
        del args, kw
        nonlocal did_ls_files
        if len(cmd) >= 2 and cmd[1] == "ls-remote":
            assert not did_ls_files
            did_ls_files = True
            data = "0123456789abcdef0123456789abcdef01234567\tHEAD"
            mock_return = mock.Mock()
            mock_return.stdout = data if encoding else data.encode("ascii")
            mock_return.stderr = "" if encoding else b""
            mock_return.returncode = 0
            return mock_return

        assert did_ls_files
        raise FileNotFoundError(errno.ENOENT, "--mocked end of test--")

    did_flock = False

    @contextlib.contextmanager
    def mock_flock(file):
        del file
        nonlocal did_flock
        assert not did_flock
        did_flock = True
        yield

    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(capellambse.helpers, "flock", mock_flock)
    caplog.set_level(logging.DEBUG)
    caplog.clear()

    with pytest.raises(FileNotFoundError, match="--mocked end of test--$"):
        capellambse.get_filehandler("git+https://domain.invalid/demo.git")

    assert did_ls_files
    assert did_flock


def test_model_loading_from_badpath_raises_FileNotFoundError():
    badpath = TEST_ROOT / "Missing.aird"
    with pytest.raises(FileNotFoundError):
        capellambse.MelodyModel(badpath)


def test_split_protocol_returns_path_objects_unchanged():
    expected = pathlib.Path.cwd()

    handler, actual = capellambse.filehandler.split_protocol(expected)

    assert handler == "file"
    assert actual == expected


@pytest.mark.parametrize(
    ["url", "expected"],
    (
        [
            ("/path/to/file", pathlib.Path("/path/to/file")),
            ("file:///path/to/file", pathlib.Path("/path/to/file")),
            ("file://localhost/path/to/file", pathlib.Path("/path/to/file")),
        ]
        if not sys.platform.startswith("win")
        else [
            ("C:/path/to/file", pathlib.Path("C:/path/to/file")),
            (r"C:\path\to\file", pathlib.Path(r"C:\path\to\file")),
            ("file:///C:/path/to/file", pathlib.Path("C:/path/to/file")),
            (
                "file://localhost/C:/path/to/file",
                pathlib.Path("C:/path/to/file"),
            ),
        ]
    ),
)
def test_split_protocol_converts_file_urls_to_paths(url, expected):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == "file"
    assert actual == expected


@pytest.mark.parametrize(
    "url",
    [
        "myhost:myrepo.git",
        "git@host:",
        "git@host:repo.git",
        "git@host:path/to/repo",
        "git@host:/path/to/repo",
    ],
)
def test_split_protocol_recognizes_scp_style_uris_as_git(url: str):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == "git"
    assert actual == url


@pytest.mark.parametrize(
    "url",
    [
        pytest.param("http://domain.invalid/path", id="http"),
        pytest.param("https://domain.invalid/path", id="https"),
        pytest.param("ftp://domain.invalid/path", id="ftp"),
        pytest.param("ftps://domain.invalid/path", id="ftps"),
        pytest.param("sftp://domain.invalid/path", id="sftp"),
        pytest.param("ssh://domain.invalid/path", id="ssh"),
    ],
)
def test_split_protocol_does_not_change_simple_urls(url: str):
    handler, actual = capellambse.filehandler.split_protocol(url)

    assert handler == url.split(":", 1)[0]
    assert actual == url


class FakeEntrypoint:
    def __init__(self, expected_name: str, expected_url: str):
        self._expected_name = expected_name
        self._expected_url = expected_url

    @property
    def name(self):
        class AlwaysEqual:
            def __eq__(_, name):  # pylint: disable=E,I
                nonlocal self
                assert name == self._expected_name
                return True

        return AlwaysEqual()

    def load(self):
        def filehandler(url: str | pathlib.Path):
            if isinstance(url, pathlib.Path):
                assert url == pathlib.Path(self._expected_url)
            else:
                assert str(url) == self._expected_url

        return filehandler

    if sys.version_info < (3, 10):

        @classmethod
        def patch(cls, monkeypatch, expected_name, expected_url):
            eps = {
                "capellambse.filehandler": (cls(expected_name, expected_url),)
            }
            monkeypatch.setattr(metadata, "entry_points", lambda: eps)

    else:

        @classmethod
        def patch(cls, monkeypatch, expected_name, expected_url):
            def entry_points(group, name):
                assert group == "capellambse.filehandler"
                assert name == expected_name
                return (cls(expected_name, expected_url),)

            monkeypatch.setattr(metadata, "entry_points", entry_points)


def test_a_single_protocol_is_not_swallowed_by_get_filehandler(
    monkeypatch,
):
    FakeEntrypoint.patch(monkeypatch, "testproto", "testproto://url")

    capellambse.get_filehandler("testproto://url")


def test_a_wrapping_protocol_separated_by_plus_is_stripped(monkeypatch):
    FakeEntrypoint.patch(monkeypatch, "realproto", "wrappedproto://url")

    capellambse.get_filehandler("realproto+wrappedproto://url")


@pytest.mark.parametrize(
    "path",
    [
        "/data/model/model.aird",
    ]
    if not sys.platform.startswith("win")
    else [
        r"S:\model\model.aird",
        r"S:/model/model.aird",
        r"\\?\S:\model\model.aird",
        r"\\modelserver\model\model.aird",
    ],
)
def test_plain_file_paths_are_recognized_as_file_protocol(monkeypatch, path):
    FakeEntrypoint.patch(monkeypatch, "file", path)

    capellambse.get_filehandler(path)


@pytest.mark.parametrize(
    "url",
    [
        "myhost:myrepo.git",
        "git@host:",
        "git@host:repo.git",
        "git@host:path/to/repo",
        "git@host:/path/to/repo",
    ],
)
def test_the_scp_short_form_is_recognized_as_git_protocol(monkeypatch, url):
    FakeEntrypoint.patch(monkeypatch, "git", url)

    capellambse.get_filehandler(url)


@pytest.mark.parametrize(
    "link",
    [
        "xtype MelodyModelTest.aird#078b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "MelodyModelTest.aird#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "xtype MelodyModel%20Test.aird#078b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "MelodyModel%20Test.aird#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
    ],
)
def test_MelodyLoader_follow_link_finds_target(link: str):
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    with pytest.raises(KeyError):
        assert loader.follow_link(None, link) is not None


@pytest.mark.parametrize(
    ["path", "subdir", "req_url"],
    [
        (
            "https://example.com/~user",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/~user/%s",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "/",
            "https://example.com/?file=demo%2Fmy%20model.aird",
        ),
        (
            "https://example.com/",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/%s",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "~user",
            "https://example.com/?file=~user%2Fdemo%2Fmy%20model.aird",
        ),
    ],
)
def test_http_file_handler_replaces_percent_s_percent_q(
    requests_mock: requests_mock.Mocker, path: str, subdir: str, req_url: str
) -> None:
    endpoint = requests_mock.get(req_url)

    file_handler = capellambse.get_filehandler(path, subdir=subdir)
    file_handler.open("demo/my model.aird", "rb").close()

    assert endpoint.called_once


def test_http_file_handler_hands_auth_to_server(
    requests_mock: requests_mock.Mocker,
) -> None:
    username = "testuser"
    password = "testpassword"
    auth_header = base64.standard_b64encode(
        f"{username}:{password}".encode("utf-8")
    ).decode("ascii")
    expected_headers = {"Authorization": f"Basic {auth_header}"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", username=username, password=password
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


def test_http_file_handlers_passed_through_custom_headers(
    requests_mock: requests_mock.Mocker,
) -> None:
    expected_headers = {"X-Test-Header": "PASSED"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", headers=expected_headers
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


def test_gitlab_artifacts_handler_uses_public_gitlab_when_no_hostname_given(
    requests_mock: requests_mock.Mocker,  # pylint: disable=unused-argument
) -> None:
    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="some-branch",
        token="my-access-token",
        job=3,
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_resolves_project_to_numeric_id(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/homegroup%2fdemo-project",
        json={"id": 1},
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project="homegroup/demo-project",
        branch="some-branch",
        token="my-access-token",
        job=3,
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_looks_up_job_id_given_a_job_name(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        json=[
            {
                "id": 3,
                "name": "make-stuff",
                "pipeline": {"ref": "somebranch"},
                "artifacts": [{"file_type": "archive"}],
            },
        ],
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="somebranch",
        token="my-access-token",
        job="make-stuff",
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


def test_gitlab_artifacts_handler_errors_if_no_job_with_artifacts_was_found(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        json=[],
    )

    with pytest.raises(RuntimeError, match="recent successful"):
        capellambse.get_filehandler(
            "glart://",
            project=1,
            branch="somebranch",
            token="my-access-token",
            job="make-stuff",
        )


def test_gitlab_artifacts_handler_handles_pagination_when_searching_for_jobs(
    requests_mock: requests_mock.Mocker,
) -> None:
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs",
        headers={
            "Link": (
                "<https://gitlab.com/api/v4/projects/1/jobs/page/2>;"
                ' rel="next"'
            )
        },
        json=[
            {
                "id": 2,
                "name": "make-stuff",
                "pipeline": {"ref": "otherbranch"},
                "artifacts": [],
            },
        ],
    )
    requests_mock.get(
        "https://gitlab.com/api/v4/projects/1/jobs/page/2",
        json=[
            {
                "id": 3,
                "name": "make-stuff",
                "pipeline": {"ref": "somebranch"},
                "artifacts": [{"file_type": "archive"}],
            },
        ],
    )

    hdl = capellambse.get_filehandler(
        "glart://",
        project=1,
        branch="somebranch",
        token="my-access-token",
        job="make-stuff",
    )

    assert isinstance(hdl, gitlab_artifacts.GitlabArtifactsFiles)
    # pylint: disable=line-too-long
    assert hdl._GitlabArtifactsFiles__path == "https://gitlab.com"  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__project == 1  # type: ignore[attr-defined]
    assert hdl._GitlabArtifactsFiles__job == 3  # type: ignore[attr-defined]


@pytest.fixture
def model_path_with_patched_version(
    request: pytest.FixtureRequest, tmp_path: pathlib.Path
) -> pathlib.Path:
    """Indirect parametrized fixture for version patched model.

    Parameters
    ----------
    request
        Special pytest fixture for access to test context, exposing the
        ``param`` attribute on indirect parametrization.
    tmp_path
        Pytest fixture giving a temporary path.

    Returns
    -------
    aird_path
        Path to Capella ``.aird`` file.
    """
    tmp_dest = tmp_path / "model"
    ignored = shutil.ignore_patterns("*.license")
    shutil.copytree(TEST_MODEL_5_0.parent, tmp_dest, ignore=ignored)
    request_param: str | tuple[str, str] = request.param
    for suffix in (".aird", ".capella"):
        model_file = (tmp_dest / TEST_MODEL).with_suffix(suffix)
        if isinstance(request_param, tuple):
            re_params = request_param
        else:
            re_params = (r"5\.0\.0", request_param)

        patched = re.sub(*re_params, model_file.read_text(encoding="utf-8"))
        model_file.write_text(patched, encoding="utf-8")
    return model_file.with_suffix(".aird")


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    ["1.3.0", "1.4.2", "6.1.0"],
    indirect=True,
)
def test_loading_model_with_unsupported_version_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginVersionError):
        capellambse.MelodyModel(model_path_with_patched_version)


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    [("http://www.eclipse.org/sirius/1.1.0", "Unknown")],
    indirect=True,
)
def test_loading_model_with_unsupported_plugin_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginError):
        capellambse.MelodyModel(model_path_with_patched_version)


def test_model_info_contains_viewpoints_and_capella_version() -> None:
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    info = loader.get_model_info()

    assert hasattr(info, "viewpoints")
    assert info.viewpoints["org.polarsys.capella.core.viewpoint"] == "5.0.0"
    assert info.capella_version == "5.0.0"


@pytest.mark.parametrize(
    ["format", "content"],
    [
        pytest.param("svg", DUMMY_SVG, id="svg"),
        pytest.param("png", DUMMY_PNG, id="png"),
    ],
)
def test_model_loads_diagrams_from_cache_by_uuid(
    tmp_path: pathlib.Path, format: str, content: bytes
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)
    dg = model.diagrams[0]
    tmp_path.joinpath(f"{dg.uuid}.{format}").write_bytes(content)

    rendered = dg.render(format)
    if isinstance(rendered, str):
        rendered = rendered.encode("utf-8")

    assert rendered == content


def test_model_will_refuse_to_render_diagrams_if_diagram_cache_was_given(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)

    with pytest.raises(RuntimeError, match="not in cache"):
        model.diagrams[0].render("svg")


def test_model_will_fall_back_to_rendering_internally_despite_the_cache_if_told_to(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(
        TEST_MODEL_5_0,
        diagram_cache=tmp_path,
        fallback_render_aird=True,
    )

    assert model.diagrams[0].render("svg")


def test_model_diagram_visible_nodes_can_be_accessed_when_a_cache_was_specified(
    tmp_path: pathlib.Path,
):
    model = capellambse.MelodyModel(TEST_MODEL_5_0, diagram_cache=tmp_path)

    assert model.diagrams[0].nodes



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Module that handles converting diagrams to the intermediary JSON format."""
from __future__ import annotations

__all__ = ["DiagramJSONEncoder"]

import collections.abc as cabc
import json
import typing as t

from capellambse import diagram

_CSSStyle = t.Union[diagram.RGB, t.Iterable[t.Union[diagram.RGB, str]], str]


class DiagramJSONEncoder(json.JSONEncoder):
    """JSON encoder that knows how to handle AIRD diagrams."""

    def default(self, o: object) -> object:
        if isinstance(o, diagram.Diagram):
            return self.__encode_diagram(o)
        if isinstance(o, diagram.Box):
            return self.__encode_box(o)
        if isinstance(o, diagram.Edge):
            return self.__encode_edge(o)
        if isinstance(o, diagram.Circle):
            return self.__encode_circle(o)
        if isinstance(o, diagram.RGB):
            return str(o)
        if isinstance(o, cabc.Sequence):
            return list(o)
        return super().default(o)

    @staticmethod
    def __encode_diagram(o: diagram.Diagram) -> object:
        return {
            "name": o.name,
            "uuid": o.uuid,
            "class": o.styleclass,
            "x": _intround(o.viewport.pos.x) if o.viewport is not None else 0,
            "y": _intround(o.viewport.pos.y) if o.viewport is not None else 0,
            "width": (
                _intround(o.viewport.size.x) if o.viewport is not None else 0
            ),
            "height": (
                _intround(o.viewport.size.y) if o.viewport is not None else 0
            ),
            "contents": [e for e in o if not e.hidden],
        }

    @staticmethod
    def __encode_box(o: diagram.Box) -> object:
        children = [
            c.uuid
            for c in o.children
            if isinstance(c, diagram.Box) and not c.port
        ]
        ports = [p.uuid for p in o.children if p.port]
        jsonobj: dict[str, object] = {
            "type": o.JSON_TYPE,
            "id": o.uuid,
            "class": o.styleclass,
            "x": _intround(o.pos.x),
            "y": _intround(o.pos.y),
            "width": _intround(o.size.x),
            "height": _intround(o.size.y),
            "context": sorted(o.context),
        }
        if o.label is not None and not o.hidelabel:
            jsonobj["label"] = _encode_label(o.label)
        if o.styleoverrides:
            jsonobj["style"] = _encode_styleoverrides(o.styleoverrides)
        if o.features:
            jsonobj["features"] = o.features
        if o.parent:
            jsonobj["parent"] = o.parent.uuid
        if children:
            jsonobj["children"] = children
        if ports:
            jsonobj["ports"] = ports

        return jsonobj

    @staticmethod
    def __encode_edge(o: diagram.Edge) -> object:
        jsonobj: dict[str, object] = {
            "type": o.JSON_TYPE,
            "id": o.uuid,
            "class": o.styleclass,
            "points": [[_intround(x), _intround(y)] for x, y in o.points],
            "labels": [_encode_label(i) for i in o.labels if not i.hidden],
        }

        if o.styleoverrides:
            jsonobj["style"] = _encode_styleoverrides(o.styleoverrides)
        return jsonobj

    @staticmethod
    def __encode_circle(o: diagram.Circle) -> object:
        jsonobj: dict[str, object] = {
            "type": o.JSON_TYPE,
            "id": o.uuid,
            "class": o.styleclass,
            "center": [_intround(p) for p in o.center],
            "radius": _intround(o.radius),
        }
        if o.styleoverrides:
            jsonobj["style"] = _encode_styleoverrides(o.styleoverrides)
        return jsonobj


def _encode_label(o: diagram.Box | str) -> object:
    if isinstance(o, str):
        return o
    return {
        "x": _intround(o.pos.x),
        "y": _intround(o.pos.y),
        "width": _intround(o.size.x),
        "height": _intround(o.size.y),
        "text": o.label,
    }


def _intround(val: float | int) -> int:
    return int(val + 0.5)


def _encode_styleoverrides(
    overrides: cabc.Mapping[str, _CSSStyle]
) -> dict[str, object]:
    return {k: _encode_style(v) for k, v in overrides.items()}


def _encode_style(style: _CSSStyle) -> object:
    if isinstance(style, diagram.RGB):
        return str(style)
    if isinstance(style, (list, tuple)):
        return [_encode_style(i) for i in style]
    return style



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = ["AttributeAuditor", "WriteProtector"]

import collections.abc as cabc
import sys
import typing as t

import capellambse

from . import helpers


class AttributeAuditor:
    """Audits access to attributes of ModelElements.

    .. warning::
        This will permanently add an audit hook to the global hook
        table. The auditor will keep the model alive, which may consume
        excessive memory. To avoid this, call the auditor object's
        ``detach()`` method once you are done with it. This is
        automatically done if you use it as a context manager.

    Examples
    --------
    >>> auditor = AttributeAuditor(model, {"name", "description"})
    >>> print(model.la.all_components[0].name)
    Hogwarts
    >>> auditor.recorded_ids
    {'0d2edb8f-fa34-4e73-89ec-fb9a63001440'}
    >>> # Cleanup
    >>> auditor.detach()

    >>> with AttributeAuditor(model, {"name", "description"}) as recorded_ids:
    ...     print(model.la.all_components[0].name)
    ...
    Hogwarts
    >>> recorded_ids
    {'0d2edb8f-fa34-4e73-89ec-fb9a63001440'}
    """

    def __init__(
        self,
        model: capellambse.MelodyModel,
        attrs: cabc.Container[str] = (),
    ) -> None:
        self.model: capellambse.MelodyModel | None = model
        self.attrs = attrs or helpers.EverythingContainer()
        self.recorded_ids: set[str] = set()

        sys.addaudithook(self.__audit)

    def __enter__(self) -> set[str]:
        return self.recorded_ids

    def __exit__(self, *_: t.Any) -> None:
        self.detach()

    def detach(self) -> None:
        self.model = None

    def __audit(self, event: str, args: tuple[t.Any, ...]) -> None:
        if event == "capellambse.getattr":
            obj, attr_name, attr_value = args
            if not hasattr(obj, "_model") or obj._model is not self.model:
                return

            if attr_name in self.attrs:
                self.recorded_ids.add(
                    attr_value if attr_name == "uuid" else obj.uuid
                )


class WriteProtector:
    """Prevents accidental modifications to a model.

    This class intentionally has very limited features. It is intended
    as inspiration and guidance for more specific classes that make more
    sophisticated use of audit events.

    This class also contains a publicly usable set of events that
    signify changes to the model.
    """

    events: t.Final = frozenset(
        {
            "capellambse.create",
            "capellambse.delete",
            "capellambse.insert",
            "capellambse.setattr",
            "capellambse.setitem",
        }
    )
    """Contains all built-in audit events that may modify the model."""

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self.__model: capellambse.MelodyModel | None = model
        sys.addaudithook(self.__audit)

    def __enter__(self) -> None:
        return None

    def __exit__(self, *_: t.Any) -> None:
        self.detach()

    def detach(self) -> None:
        self.__model = None

    def __audit(self, event: str, args: tuple[t.Any, ...]) -> None:
        if self.__model is not None and event in self.events:
            obj: capellambse.model.GenericElement = args[0]
            if obj._model is self.__model:
                raise RuntimeError(
                    f"Cannot modify the model {self.__model.info}"
                )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Classes that represent different property value types."""

from __future__ import annotations

import abc
import collections.abc as cabc

import capellambse._namespaces as _n

from . import model, validation
from .core import AttributeProperty, Generic, XMLDictProxy

NAMESPACED_PV = "org.polarsys.capella.core.data.capellacore:{}"
XTYPE_KEY = f"{{{_n.NAMESPACES['xsi']}}}type"


class GenericPropertyValue(Generic, metaclass=abc.ABCMeta):
    """Base class for property value types."""

    description = AttributeProperty("xml_element", "description")
    xtype = AttributeProperty(
        "xml_element", f"{{{_n.NAMESPACES['xsi']}}}type", writable=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = None

    @staticmethod
    def applyto(pvmt_ext, defelem, modelobj, targetelem):
        """Apply a property value to ``targetelem``.

        Parameters
        ----------
        pvmt_ext
            The PVMT Extension object.
        defelem
            The ``ownedPropertyValues`` element that should be applied.
        modelobj
            A model object's XML element that will be a (direct or
            indirect) parent to this property value.
        targetelem
            The new ``ownedPropertyValues`` element.
        """
        for attr in (f"{{{_n.NAMESPACES['xsi']}}}type", "name", "value"):
            try:
                targetelem.attrib[attr] = defelem.attrib[attr]
            except KeyError:
                pass

        targetelem.attrib.update(
            {
                "id": pvmt_ext.model.generate_uuid(modelobj),
                "appliedPropertyValues": pvmt_ext.model.create_link(
                    modelobj, defelem
                ),
            }
        )

    @staticmethod
    @abc.abstractmethod
    def cast(value):
        """Cast the given string into an appropriate Python object."""

    def serialize(self, value, element=None):
        """Serialize the given value into an XML string.

        This function must be able to handle two types of input values:

        1. The type produced by ``.cast()``, which shall be turned back
           into its XML form, so that it can be passed into ``.cast()``
           again.
        2. The XML representation of a valid value, i.e. its own output.

        If it is ambiguous which of the two types is being handled, the
        former shall be assumed.

        Note that escaping of XML special characters is handled by the
        underlying XML library.

        The default implementation works for the simple case where
        ``cast`` is a type constructor, but for more complex cases it
        should be overridden.

        Parameters
        ----------
        value
            The value that should be serialized.
        element
            The XML element into which the value will be inserted. This
            may be used to construct links across fragment boundaries.
            This parameter may be None, in which case it is assumed that
            all elements exist within the same fragment.
        """
        del element

        if not isinstance(self.cast, type):
            type_name = type(self).__name__
            raise TypeError(
                f"{type_name}.cast is not a simple type;"
                f"please implement {type_name}.serialize"
            )
        # pylint: disable=isinstance-second-argument-not-valid-type
        if isinstance(value, self.cast):
            return str(value)
        return str(self.cast(value))


class StringPropertyValue(GenericPropertyValue):
    """A string property value."""

    cast = str


class BooleanPropertyValue(GenericPropertyValue):
    """A boolean property value."""

    @staticmethod
    def cast(value):
        if value == "true":
            return True
        if value == "false":
            return False
        raise ValueError(
            f"Boolean values must be either 'true' or 'false', not {value!r}"
        )

    @classmethod
    def serialize(cls, value, element=None):
        del element
        if isinstance(value, str):
            value = cls.cast(value)
        return "true" if value else "false"


class IntegerPropertyValue(GenericPropertyValue):
    """An integer property value."""

    cast = int

    @property
    def unit(self):
        """Return the measurement unit of this property value."""
        unit = self.xml_element.xpath('*[@name="__UNIT__"]/@value')
        return unit[0] if unit else None


class FloatPropertyValue(GenericPropertyValue):
    """A floating point property value."""

    cast = float

    @property
    def unit(self):
        """Return the measurement unit of this property value."""
        unit = self.xml_element.xpath('*[@name="__UNIT__"]/@value')
        return unit[0] if unit else None


class EnumerationPropertyValue(GenericPropertyValue):
    """An enumeration property value."""

    def __init__(self, *args, typedef=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.typedef = typedef

    @property  # FIXME: what's the type?
    def parent(self):
        """Return the parent group of this property value."""
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent
        self._typedef = None

    @property
    def typedef(self):
        """Return the type definition of this enumeration."""
        if self._typedef is None:
            if self.parent is None:
                raise TypeError("Cannot resolve typedef: No parent")
            domain = self.parent.parent
            assert isinstance(domain, model.Domain)
            self._typedef = domain.enums[
                self.xml_element.attrib["type"].split("#")[-1]
            ]
        return self._typedef

    @typedef.setter
    def typedef(self, typedef):
        self._typedef = typedef

    @property
    def default_value(self):
        """Return this property's default value."""
        return self.typedef[self.xml_element.attrib["value"].split("#")[-1]]

    @staticmethod
    def applyto(pvmt_ext, defelem, modelobj, targetelem):
        super().applyto(pvmt_ext, defelem, modelobj, targetelem)

        typeelem = pvmt_ext.model[defelem.attrib["type"].split("#")[-1]]
        valelem = pvmt_ext.model[defelem.attrib["value"].split("#")[-1]]
        targetelem.attrib.update(
            {
                "type": pvmt_ext.model.create_link(modelobj, typeelem),
                "value": pvmt_ext.model.create_link(modelobj, valelem),
            }
        )

    def cast(self, value):
        return self.typedef[value.split("#")[-1]]["name"]

    def serialize(self, value, element=None):
        if not isinstance(value, str):
            raise TypeError("Enumeration values must be str")

        if element is None:
            element = self.xml_element

        pvext = self.parent.parent.parent

        for typeitem in self.typedef.items():
            if value == typeitem[1]["name"]:
                return pvext.model.create_link(
                    element, pvext.model[typeitem[0]]
                )

        value = value.split("#")[-1]
        if value in self.typedef:
            # It's already the UUID. Normalize it.
            return pvext.model.create_link(element, pvext.model[value])

        raise ValueError(f"Not a valid value for this enumeration: {value}")


class EnumerationPropertyType(XMLDictProxy):
    """Maps the literals' UUIDs to their human-readable values."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, **kwargs, childtag="ownedLiterals", keyattr="id"
        )

    @property
    def literals(self):
        """Return a list of valid literal values for this enumeration."""
        return list(self.values())

    def _extract_value(self, element):
        return dict(element.attrib)

    def _insert_value(self, element, value):
        element.attrib.update(value)


PROPERTY_LOADER: cabc.Mapping[str, type[GenericPropertyValue]] = {
    "EnumerationPropertyValue": EnumerationPropertyValue,
    "StringPropertyValue": StringPropertyValue,
    "FloatPropertyValue": FloatPropertyValue,
    "BooleanPropertyValue": BooleanPropertyValue,
    "IntegerPropertyValue": IntegerPropertyValue,
}


def select_property_loader(element):
    """Execute the appropriate loader for the PV definition element."""
    xtype = str(element.attrib[XTYPE_KEY]).rsplit(":", maxsplit=1)[-1]
    if xtype not in PROPERTY_LOADER:
        raise NotImplementedError(f"XSI type {xtype} is not yet supported")
    return PROPERTY_LOADER[xtype](element)


class AppliedPropertyValueGroup(XMLDictProxy):
    """A group of applied property values."""

    def __init__(self, pvmt_ext, *args, **kwargs):
        """Create an AppliedPropertyValueGroup.

        Parameters
        ----------
        pvmt_ext
            The :class:`PVMTExtension` object of the model.
        """
        super().__init__(
            *args, childtag="ownedPropertyValues", keyattr="name", **kwargs
        )
        self._groupdef = self._get_named_pv_group(
            pvmt_ext, self.xml_element.attrib["name"]
        )
        self._pvmt_ext = pvmt_ext

    def get_definition(self, key):
        """Return the PV definition instance for the given key."""
        for propdef in self._groupdef.values():
            if propdef.name == key:
                return propdef
        raise KeyError(f"Key {key} is not defined in the group")

    def _extract_value(self, element):
        pvtype = self._groupdef[
            element.attrib["appliedPropertyValues"].split("#")[-1]
        ]
        if "value" in element.attrib:
            return pvtype.cast(element.attrib["value"])
        else:
            return None

    def _prepare_child(self, element, key):
        super()._prepare_child(element, key)
        propdef = self.get_definition(key)
        propdef.applyto(
            self._pvmt_ext,
            propdef.xml_element,
            self._groupdef.xml_element,
            element,
        )

    def _insert_value(self, element, value):
        pvtype = self._groupdef[
            element.attrib["appliedPropertyValues"].split("#")[-1]
        ]
        element.set("value", pvtype.serialize(value, element))

    @classmethod
    def applyto(cls, pvmt_ext, xml_element, groupname):
        """Apply the named property value group to the given element.

        Parameters
        ----------
        pvmt_ext
            The PVMT extension object
        xml_element
            The XML element of the target object
        groupname
            The fully qualified name of the PVMT group

        Returns
        -------
        element
            The newly created XML element, a child of xml_element.
        """
        groupdef = cls._get_named_pv_group(pvmt_ext, groupname)
        validation.validate_group_scope(pvmt_ext, groupdef, xml_element)

        groupelem = xml_element.makeelement(
            "ownedPropertyValueGroups",
            attrib={
                f"{{{_n.NAMESPACES['xsi']}}}type": NAMESPACED_PV.format(
                    "PropertyValueGroup"
                ),
                "id": pvmt_ext.model.generate_uuid(xml_element),
                "name": groupname,
                "appliedPropertyValueGroups": pvmt_ext.model.create_link(
                    xml_element, groupdef.xml_element
                ),
            },
        )

        # Apply the group's property values
        for defelem in groupdef.xml_element.iterchildren(
            "ownedPropertyValues"
        ):
            propelem = groupelem.makeelement("ownedPropertyValues")
            PROPERTY_LOADER[
                defelem.get(f"{{{_n.NAMESPACES['xsi']}}}type").split(":")[-1]
            ].applyto(pvmt_ext, defelem, xml_element, propelem)
            groupelem.append(propelem)

        xml_element.append(groupelem)
        pvmt_ext.model.idcache_index(groupelem)
        if "appliedPropertyValueGroups" in xml_element.attrib:
            xml_element.attrib["appliedPropertyValueGroups"] = " ".join(
                [
                    xml_element.attrib["appliedPropertyValueGroups"],
                    f'#{groupelem.attrib["id"]}',
                ]
            )
        else:
            xml_element.attrib[
                "appliedPropertyValueGroups"
            ] = f'#{groupelem.attrib["id"]}'
        return groupelem

    @staticmethod
    def _get_named_pv_group(pvmt_ext, groupname):
        domainname, groupname = groupname.split(".")
        domain = [d for d in pvmt_ext.values() if d.name == domainname]
        if len(domain) != 1:
            raise ValueError(f"Domain {domainname!r} not found in model")

        group = [g for g in domain[0].values() if g.name == groupname]
        if len(group) != 1:
            raise ValueError(
                f"Group {groupname!r} not found in model domain {domainname!r}"
            )

        return group[0]



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
import logging
import pathlib

import cssutils
import pytest
from lxml import etree

import capellambse
import capellambse.diagram
from capellambse.svg import (
    SVGDiagram,
    decorations,
    generate,
    helpers,
    style,
    symbols,
)

cssutils.log.setLevel(logging.CRITICAL)

TEST_LAB = "[LAB] Wizzard Education"
TEST_DIAGS = [
    TEST_LAB,
    "[OAB] Operational Context",
    "[OCB] Operational Capabilities",
    "[OPD] Obtain food via hunting",
    "[MSM] States of Functional Human Being",
    "[SAB] System",
    "[OEBD] Operational Context",
    "[PAB] Physical System",
    "[LDFB] Test flow",
    "[CC] Capability",
]
TEST_DECO = set(style.STATIC_DECORATIONS.keys()) - {"__GLOBAL__"}
FREE_SYMBOLS = {
    "OperationalCapabilitySymbol",
    "AndControlNodeSymbol",
    "ItControlNodeSymbol",
    "OrControlNodeSymbol",
    "FinalStateSymbol",
    "InitialPseudoStateSymbol",
    "TerminatePseudoStateSymbol",
    "StickFigureSymbol",
}


@pytest.fixture(name="tmp_json")
def tmp_json_fixture(
    model: capellambse.MelodyModel, tmp_path: pathlib.Path
) -> pathlib.Path:
    """Return tmp path of diagram json file."""
    dest = tmp_path / (TEST_LAB + ".json")
    diagram = model.diagrams.by_name(TEST_LAB).render(None)
    diagram_json = capellambse.diagram.DiagramJSONEncoder().encode(diagram)
    dest.write_text(diagram_json)
    return dest


class TestSVG:
    def test_diagram_meta_data_attributes(
        self, tmp_json: pathlib.Path
    ) -> None:
        diag_meta = generate.DiagramMetadata.from_dict(
            json.loads(tmp_json.read_text())
        )
        assert diag_meta.name == TEST_LAB
        assert diag_meta.pos == (15, 15)
        assert diag_meta.size == (1162, 611)
        assert diag_meta.viewbox == "15 15 1162 611"
        assert diag_meta.class_ == "Logical Architecture Blank"

    def test_diagram_from_json_path_componentports(
        self, tmp_json: pathlib.Path
    ) -> None:
        tree = etree.fromstring(
            SVGDiagram.from_json_path(tmp_json).to_string()
        )

        cp_in_exists: bool = False
        cp_inout_exists: bool = False
        cp_out_exists: bool = False
        cp_unset_exists: bool = False
        cp_reference_exists: bool = False

        for item in tree.iter():
            # The class CP should not exist anymore as it has been replaced
            # with CP_IN, CP_OUT, CP_UNSET or CP_INOUT
            assert item.get("class") != "Box CP"

            # Check that the classes CP_IN, CP_OUT, CP_UNSET and CP_INOUT exist
            if item.get("class") == "Box CP_IN":
                cp_in_exists = True
            elif item.get("class") == "Box CP_OUT":
                cp_out_exists = True
            elif item.get("class") == "Box CP_INOUT":
                cp_inout_exists = True
            elif item.get("class") == "Box CP_UNSET":
                cp_unset_exists = True

            # Check that reference symbol for CP exists
            if (
                item.tag == "{http://www.w3.org/2000/svg}symbol"
                and item.get("id") == "ComponentPortSymbol"
            ):
                cp_reference_exists = True

        assert cp_in_exists
        assert cp_out_exists
        assert cp_inout_exists
        assert cp_unset_exists
        assert cp_reference_exists

    @pytest.fixture
    def tmp_svg(self, tmp_path: pathlib.Path) -> SVGDiagram:
        name = "Test svg"
        meta = generate.DiagramMetadata(
            pos=(0, 0), size=(1, 1), name=name, class_="TEST"
        )
        svg = SVGDiagram(meta, [])
        svg.drawing.filename = str(tmp_path / name)
        return svg

    def test_diagram_saves(self, tmp_svg: SVGDiagram) -> None:
        tmp_svg.save_drawing()
        assert pathlib.Path(tmp_svg.drawing.filename).is_file()

    # FIXME: change this to a parametrized test, do not use if- or
    # for-statements in a unit test
    def test_css_colors(self, tmp_json: pathlib.Path) -> None:
        COLORS_TO_CHECK = {
            ".LogicalArchitectureBlank g.Box.CP_IN > line": {
                "stroke": "#000000"
            },
            ".LogicalArchitectureBlank g.Box.CP_IN > rect,"
            " .LogicalArchitectureBlank g.Box.CP_IN > use": {
                "fill": "#FFFFFF",
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.CP_OUT > line": {
                "stroke": "#000000"
            },
            ".LogicalArchitectureBlank g.Box.CP_OUT > rect,"
            " .LogicalArchitectureBlank g.Box.CP_OUT > use": {
                "fill": "#FFFFFF",
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.CP_INOUT > line": {
                "stroke": "#000000,"
            },
            ".LogicalArchitectureBlank g.Box.CP_INOUT > rect,"
            " .LogicalArchitectureBlank g.Box.CP_INOUT > use": {
                "fill": "#FFFFFF",
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Edge > path": {
                "fill": "none",
                "stroke": "rgb(0, 0, 0)",
            },
            ".LogicalArchitectureBlank g.Box > line": {"stroke": "#000000"},
            ".LogicalArchitectureBlank g.Box > rect,"
            " .LogicalArchitectureBlank g.Box > use": {
                "fill": "transparent",
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.Annotation > line": {
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Box.Annotation > rect,"
            " .LogicalArchitectureBlank g.Box.Annotation > use": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Box.Constraint > line": {
                "stroke": "#888888",
            },
            ".LogicalArchitectureBlank g.Box.Constraint > rect,"
            " .LogicalArchitectureBlank g.Box.Constraint > use": {
                "fill": "#FFF5B5",
                "stroke": "#888888",
            },
            ".LogicalArchitectureBlank g.Box.Constraint > text": {
                "fill": "#000000"
            },
            ".LogicalArchitectureBlank g.Box.Note > line": {
                "stroke": "#FFCC66"
            },
            ".LogicalArchitectureBlank g.Box.Note > rect,"
            " .LogicalArchitectureBlank g.Box.Note > use": {
                "fill": " #FFFFCB",
                "stroke": " #FFCC66",
            },
            ".LogicalArchitectureBlank g.Box.Note > text": {
                "fill": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.Requirement > line": {
                "stroke": "#72496E",
            },
            ".LogicalArchitectureBlank g.Box.Requirement > rect,"
            " .LogicalArchitectureBlank g.Box.Requirement > use": {
                "fill": "#D9C4D7",
                "stroke": "#72496E",
            },
            ".LogicalArchitectureBlank g.Box.Requirement > text": {
                "fill": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.Text > line": {
                "stroke": "transparent",
            },
            ".LogicalArchitectureBlank g.Box.Text > rect,"
            " .LogicalArchitectureBlank g.Box.Text > use": {
                "stroke": "transparent",
            },
            ".LogicalArchitectureBlank g.Edge > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle > circle": {
                "fill": "#000000",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.Connector > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.Connector > circle": {
                "fill": "#B0B0B0",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.Connector > path": {
                "stroke": "#B0B0B0",
            },
            ".LogicalArchitectureBlank g.Edge.Constraint > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.Constraint > circle": {
                "fill": "#000000",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.Constraint > path": {
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Edge.Note > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.Note > circle": {
                "fill": "#000000",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.Note > path": {
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Edge.RequirementRelation > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.RequirementRelation"
            " > circle": {
                "fill": "#72496E",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.RequirementRelation > path": {
                "stroke": "#72496E"
            },
            ".LogicalArchitectureBlank g.Edge.RequirementRelation > text": {
                "fill": "#72496E",
            },
            ".LogicalArchitectureBlank g.Box.CP > line": {
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.CP > rect,"
            " .LogicalArchitectureBlank g.Box.CP > use": {
                "fill": "#FFFFFF",
                "stroke": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.FIP > rect,"
            " .LogicalArchitectureBlank g.Box.FIP > use": {
                "fill": "#E08503",
            },
            ".LogicalArchitectureBlank g.Box.FOP > rect,"
            " .LogicalArchitectureBlank g.Box.FOP > use": {
                "fill": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Box.LogicalActor > line": {
                "stroke": "#4A4A97",
            },
            ".LogicalArchitectureBlank g.Box.LogicalActor > text": {
                "fill": "#000000",
            },
            ".LogicalArchitectureBlank g.Box.LogicalComponent > line": {
                "stroke": "#4A4A97",
            },
            ".LogicalArchitectureBlank g.Box.LogicalComponent > text": {
                "fill": "#4A4A97",
            },
            ".LogicalArchitectureBlank g.Box.LogicalFunction > line": {
                "stroke": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Box.LogicalFunction > rect,"
            " .LogicalArchitectureBlank g.Box.LogicalFunction > use": {
                "fill": "#C5FFA6",
                "stroke": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Box.LogicalFunction > text": {
                "fill": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Edge.FunctionalExchange > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.FunctionalExchange > circle": {
                "fill": "#095C2E",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.FunctionalExchange > path": {
                "stroke": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Edge.FunctionalExchange > text": {
                "fill": "#095C2E",
            },
            ".LogicalArchitectureBlank g.Edge.ComponentExchange > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.ComponentExchange > circle": {
                "fill": "#4A4A97",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.ComponentExchange > path": {
                "stroke": "#4A4A97",
            },
            ".LogicalArchitectureBlank g.Edge.ComponentExchange > text": {
                "fill": "#4A4A97",
            },
            ".LogicalArchitectureBlank g.Edge.FIPAllocation > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.FIPAllocation > circle": {
                "fill": "#E08503",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.FIPAllocation > path": {
                "stroke": "#E08503",
            },
            ".LogicalArchitectureBlank g.Edge.FOPAllocation > rect": {
                "fill": "none",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Circle.FOPAllocation > circle": {
                "fill": "#095C2E",
                "stroke": "none",
            },
            ".LogicalArchitectureBlank g.Edge.FOPAllocation > path": {
                "stroke": "#095C2E",
            },
        }

        tree = etree.fromstring(
            SVGDiagram.from_json_path(tmp_json).to_string()
        )
        style_ = tree.xpath(
            "/x:svg/x:defs/x:style",
            namespaces={"x": "http://www.w3.org/2000/svg"},
        )[0]

        stylesheet = cssutils.parseString(style_.text)
        for rule in stylesheet:
            for element, prop in COLORS_TO_CHECK.items():
                if (
                    rule.type == rule.STYLE_RULE
                    and rule.selectorText == element
                ):
                    for key, val in prop.items():
                        try:
                            property_value = rule.style.getProperty(
                                key
                            ).propertyValue.value
                        except AttributeError:
                            # FIXME: rules are duplicated with different
                            # values -> should be merged first
                            print(f"Missing attribute {key}")
                            continue
                        if val == "none":
                            assert property_value == "none"
                        elif key in ["fill", "stroke"]:
                            assert (
                                property_value
                                == cssutils.css.ColorValue(val).value
                            )
                        else:
                            raise NotImplementedError

    @pytest.mark.parametrize("diagram_name", TEST_DIAGS)
    def test_diagram_decorations(
        self, model: capellambse.MelodyModel, diagram_name: str
    ):
        """Test diagrams get rendered successfully."""
        diag = model.diagrams.by_name(diagram_name)
        diag.render("svg")


class TestSVGStyling:
    LAB = "Logical Architecture Blank"
    FCD = "Functional Chain Description"
    OEB = "Operational Entity Blank"
    FEX = "FunctionalExchange"
    OEX = "OperationalExchange"
    BOX = "LogicalComponentSymbol"

    class TestSVGStylesheet:
        def test_svg_stylesheet_as_str(self, tmp_json) -> None:
            svg = SVGDiagram.from_json_path(tmp_json)
            for line in str(svg.drawing.stylesheet).splitlines():
                assert line.startswith(".LogicalArchitectureBlank")

        def test_svg_stylesheet_builder_fails_when_no_class_was_given(self):
            expected_error_msg = (
                "Invalid type for class_ 'NoneType'. This needs to be a str."
            )

            with pytest.raises(TypeError) as error:
                style.SVGStylesheet(None)  # type: ignore[arg-type]

            assert error.value.args[0] == expected_error_msg

    @pytest.mark.parametrize(
        "style_,diagstyle,expected",
        [
            pytest.param(None, "None", None, id="No modify"),
            pytest.param(FEX, "None", None, id="Unregistered diagramstyle"),
            pytest.param(BOX, LAB, None, id="Found in deco"),
            pytest.param(FEX, FCD, None, id="Layer is prefix"),
            pytest.param(FEX, OEB, OEX, id="Fixed FExchange styleclass"),
        ],
    )
    def test_get_symbol_styleclass(
        self, style_: str | None, diagstyle: str, expected: str | None
    ) -> None:
        assert style.get_symbol_styleclass(style_, diagstyle) == expected


class TestDecoFactory:
    @pytest.mark.parametrize(
        "class_",
        ["LogicalComponentSymbol", "LogicalHumanActorSymbol", "EntitySymbol"],
    )
    def test_deco_factory_contains_styling_for_given_styleclass(
        self, class_: str
    ):
        assert class_ in decorations.deco_factories

    def test_deco_factory_returns_symbol_factory_for_given_styleclass(self):
        assert decorations.deco_factories["PortSymbol"] is symbols.port_symbol

    @pytest.mark.parametrize(
        "class_", ["ImaginaryClassSymbol", "NothingSymbol"]
    )
    def test_deco_factory_logs_error_when_not_containing_given_styleclass(
        self, caplog, class_: str
    ):
        assert class_ not in decorations.deco_factories
        assert (
            decorations.deco_factories[class_]
            is decorations.deco_factories["ErrorSymbol"]
        )
        with caplog.at_level(0, logger="decorations"):
            assert (
                caplog.messages[-1] == f"{class_} wasn't found in factories."
            )

    @pytest.mark.parametrize("attr", ["start", "translate", "offsets"])
    def test_making_linear_gradient_faulty_cases_raise_ValueError(
        self, attr: str
    ):
        params = {
            "id_": "test",
            attr: (0, 0, 0),
        }
        with pytest.raises(ValueError):
            symbols._make_lgradient(**params)

    def test_making_linear_gradient_with_translate(self):
        gradient = symbols._make_lgradient("test", translate=(0, 0))
        assert gradient.attribs.get("gradientTransform") == "translate(0 0)"


class TestSVGHelpers:
    def test_check_for_horizontal_overflow_recognizes_tabs_and_breaks(
        self,
    ) -> None:
        lines, margin, max_text_width = helpers.check_for_horizontal_overflow(
            "             • item 1\n             • item 2", 100, 0, 0
        )
        assert lines == ["             • item 1", "             • item 2"]
        assert 10 <= margin < 13
        for line in lines:
            assert capellambse.helpers.extent_func(line)[0] <= max_text_width



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Stylesheet generator for SVG diagrams."""
from __future__ import annotations

import collections.abc as cabc
import io
import itertools
import logging
import operator
import re
import textwrap
import typing as t

from lxml import etree
from svgwrite import base, gradients

from capellambse import diagram

if t.TYPE_CHECKING:
    from .drawing import Drawing

logger = logging.getLogger(__name__)
RE_ELMCLASS = re.compile(r"^([A-Z][a-z_]*)(\.[A-Za-z][A-Za-z0-9_]*)?(:.+)?$")
CUSTOM_STYLE_ATTRS = {"marker-fill"}

# TODO refactor to dynamically determine needed decorations
STATIC_DECORATIONS: dict[str, tuple[str, ...]] = {
    "__GLOBAL__": (
        "ErrorSymbol",
        "RequirementSymbol",
    ),
    "Error": (),
    "Class Diagram Blank": ("ClassSymbol",),
    "Functional Chain Description": (
        "AndControlNodeSymbol",
        "IterateControlNodeSymbol",
        "OrControlNodeSymbol",
        "FunctionalExchangeSymbol",
        "FunctionSymbol",
    ),
    "Logical Architecture Blank": (
        "ComponentPortSymbol",
        "LogicalActorSymbol",
        "LogicalComponentSymbol",
        "LogicalFunctionSymbol",
        "LogicalHumanActorSymbol",
        "LogicalHumanComponentSymbol",
        "PortSymbol",
        "StickFigureSymbol",
        "FunctionalExchangeSymbol",
        "ComponentExchangeSymbol",
    ),
    "Logical Data Flow Blank": (
        "LogicalFunctionSymbol",
        "PortSymbol",
        "FunctionalExchangeSymbol",
    ),
    "Missions Capabilities Blank": (
        "CapabilitySymbol",
        "MissionSymbol",
        "SystemActorSymbol",
        "SystemComponentSymbol",
        "SystemHumanActorSymbol",
    ),
    "Mode State Machine": (
        "FinalStateSymbol",
        "InitialPseudoStateSymbol",
        "ModeSymbol",
        "StateSymbol",
        "TerminatePseudoStateSymbol",
    ),
    "Operational Capabilities Blank": (
        "EntitySymbol",
        "OperationalActorBoxSymbol",
        "OperationalCapabilitySymbol",
    ),
    "Operational Entity Blank": (
        "EntitySymbol",
        "OperationalActivitySymbol",
        "OperationalActorBoxSymbol",
        "OperationalExchangeSymbol",
    ),
    "Operational Entity Breakdown": ("OperationalActorSymbol", "EntitySymbol"),
    "Operational Process Description": (
        "AndControlNodeSymbol",
        "IterateControlNodeSymbol",
        "OperationalActivitySymbol",
        "OrControlNodeSymbol",
        "OperationalExchangeSymbol",
    ),
    "Operational Activity Interaction Blank": (
        "OperationalActivitySymbol",
        "OperationalExchangeSymbol",
    ),
    "Physical Architecture Blank": (
        "PhysicalLinkSymbol",
        "ComponentExchangeSymbol",
        "ComponentPortSymbol",
        "PortSymbol",
    ),
    "Physical Data Flow Blank": (
        "FunctionalExchangeSymbol",
        "PhysicalFunctionSymbol",
        "PortSymbol",
    ),
    "System Architecture Blank": (
        "ComponentExchangeSymbol",
        "ComponentPortSymbol",
        "FunctionalExchangeSymbol",
        "PhysicalLinkSymbol",
        "SystemActorSymbol",
        "SystemComponentSymbol",
        "SystemFunctionSymbol",
        "SystemHumanActorSymbol",
        "PortSymbol",
    ),
    "System Data Flow Blank": (
        "FunctionalExchangeSymbol",
        "PortSymbol",
        "SystemFunctionSymbol",
    ),
    "Contextual Capability": (
        "CapabilitySymbol",
        "MissionSymbol",
        "SystemActorSymbol",
        "SystemHumanActorSymbol",
    ),
}
MODIFY_STYLECLASS = {"FunctionalExchange"}


def get_symbol_styleclass(style: str | None, dstyle: str | None) -> str | None:
    if (
        style not in MODIFY_STYLECLASS
        or dstyle not in STATIC_DECORATIONS
        or style in STATIC_DECORATIONS[dstyle]
    ):
        return None

    capitals = (dstyle or "").split(" ", maxsplit=1)
    assert capitals
    layer = capitals[0]
    if style.startswith(layer):
        return None

    scapitals = re.findall(r"[A-Z][^A-Z]*", style)
    symbol = f'{layer}{"".join(scapitals[1:])}'
    if f"{symbol}Symbol" in STATIC_DECORATIONS[dstyle]:
        return symbol
    return None


class Styling:
    """Container for style attributes of svg objects.

    Notes
    -----
    Attributes containing '-' are only referenceable via getattr() or
    subscripting syntax, due to Python identifier naming rules.
    """

    _marker: bool = False

    def __init__(
        self, diagram_class: str | None, class_: str, prefix: str = "", **attr
    ):
        self._diagram_class = diagram_class
        self._class = class_
        self._prefix = prefix

        for key, val in attr.items():
            setattr(self, key, val)

    def __setattr__(self, name: str, value: t.Any) -> None:
        if not name.startswith("_"):
            name = name.replace("_", "-")
        super().__setattr__(name, value)

    def __getattribute__(self, attr: str) -> str:
        if attr in {"marker-start", "marker-end"}:
            defaultstyles = diagram.get_style(self._diagram_class, self._class)
            try:
                value = super().__getattribute__(attr)
            except AttributeError as err:
                try:
                    value = defaultstyles[self._style_name(attr)]
                except KeyError:
                    raise err from None
            try:
                stroke = self.stroke
            except AttributeError:
                stroke = (
                    defaultstyles.get(self._style_name("stroke")) or "#000"
                )
            return f'url("#{value}_{diagram.RGB.fromcss(stroke).tohex()}")'

        return super().__getattribute__(attr)

    def __bool__(self) -> bool:
        try:
            next(iter(self))
        except StopIteration:
            return False
        return True

    def __iter__(self) -> cabc.Iterator[str]:
        defaultstyles = diagram.get_style(self._diagram_class, self._class)
        for attr in ("marker-start", "marker-end"):
            if (
                not self._marker
                and not getattr(super(), attr, None)
                and self._style_name(attr) in defaultstyles
            ):
                yield attr

        yield from itertools.filterfalse(
            operator.methodcaller("startswith", "_"), dir(self)
        )

    def __getitem__(self, attrs: str | cabc.Iterable[str]) -> str | None:
        if isinstance(attrs, str):
            attrs = (attrs,) if attrs else self
        return (
            "; ".join(f"{a}: {self._to_css(getattr(self, a))}" for a in attrs)
            or None
        )

    @classmethod
    def _to_css(
        cls, value: float | int | str | diagram.RGB | cabc.Iterable | None
    ) -> float | int | str:
        if isinstance(value, (str, int, float)):
            return value
        elif value is None:
            return "none"
        elif isinstance(value, diagram.RGB):
            return f"#{value.tohex()}"
        elif isinstance(value, cabc.Iterable):
            return f'url("#{cls._generate_id("CustomGradient", value)}")'
        raise ValueError(f"Invalid styling value: {value!r}")

    @staticmethod
    def _generate_id(
        name: str, value: cabc.Iterable[str | diagram.RGB]
    ) -> str:
        """Return unqiue identifier for given css-value."""
        return "_".join(
            itertools.chain(
                (name,),
                (diagram.RGB.fromcss(v).tohex() for v in value),
            ),
        )

    def __str__(self) -> str:
        return self[""] or ""

    def _style_name(self, attr: str) -> str:
        if not self._prefix:
            return attr

        return "_".join((self._prefix, attr))


class Style(base.BaseElement):
    """An embedded cascading style sheet."""

    elementname = "style"

    def __init__(self, text: str, **extra: dict[str, t.Any]) -> None:
        """Initialize Style class.

        Parameters
        ----------
        text
            The stylesheet
        extra
            Additional attributes as keyword arguments
        """
        super().__init__(**extra)
        self.text = text

    def get_xml(self) -> etree._Element:
        xml = super().get_xml()
        xml.text = self.text
        return xml


class SVGStylesheet:
    """CSS stylesheet for SVG."""

    def __init__(self, class_: str):
        if not isinstance(class_, str):
            raise TypeError(
                f"Invalid type for class_ '{type(class_).__name__}'. This"
                " needs to be a str."
            )

        self.drawing_class = class_
        self.builder = StyleBuilder(class_)
        self.create()

    @property
    def sheet(self) -> Style:
        """Return Sheet str from builder."""
        return Style(text=self.builder.sheet.getvalue())

    def create(self) -> None:
        """Generate a stylesheet for the given drawing."""
        self.builder.write_styles()
        self.static_deco = STATIC_DECORATIONS[
            "__GLOBAL__"
        ] + STATIC_DECORATIONS.get(self.drawing_class or "", ())

    def yield_gradients(self) -> cabc.Iterator[gradients.LinearGradient]:
        """Yield an svgwrite LinearGradient for all gradients in this sheet."""
        for gradname, stopcolors in self.builder.gradients.items():
            grad = gradients.LinearGradient(
                id_=gradname, start=("0%", "0%"), end=("0%", "100%")
            )
            for offset, color in zip((0, 1), stopcolors):
                grad.add_stop_color(offset=offset, color=color)

            yield grad

    def __str__(self) -> str:
        return self.sheet.text


class StyleBuilder:
    """Helper class that bundles together all needed objects."""

    _base = textwrap.dedent(
        """\
        {cls}* {{ shape-rendering: geometricPrecision; }}
        {cls}text {{ font-family: "Segoe UI"; font-size: 8pt; }}
        {cls}g {{ cursor: pointer; }}
        {cls}g.Edge > path {{ fill: none; stroke: rgb(0, 0, 0); }}
        """
    )
    _highlight_on_hover = {
        "ComponentExchange": diagram.RGB(8, 138, 189),
        "ExchangeItemElement": diagram.RGB(0, 0, 0),
        "FIPAllocation": diagram.RGB(255, 0, 0),
        "FOPAllocation": diagram.RGB(255, 0, 0),
        "FunctionalExchange": diagram.RGB(0, 0, 255),
        "PhysicalLink": diagram.RGB(239, 41, 41),
    }

    sheet: io.StringIO
    styles: dict[str, dict[str, diagram.CSSdef]]

    def __init__(self, class_: str | None):
        self.class_ = class_
        self.sheetclass = re.sub(r"\s+", "", class_ or "")
        self.stylewriters = {
            "Box": self._write_styles_box,
            "Edge": self._write_styles_edge,
        }
        self.gradients: dict[str, str | tuple[str, ...]] = {}
        self.create()

    def write_styles(self) -> None:
        """Write edge and box styles to sheet."""
        for key, styles in self.styles.items():
            if not styles:  # pragma: no cover
                continue
            elmtype_match = RE_ELMCLASS.match(key)
            if elmtype_match is None:  # pragma: no cover
                logger.error("Invalid style key: %s", key)
                continue

            elmtype, elmclass, pseudoclass = elmtype_match.groups()
            if elmtype not in self.stylewriters:
                continue
            write_styles_real = self.stylewriters[elmtype]
            write_styles_real(elmclass or "", pseudoclass or "", styles)

    def create(self) -> None:
        """Create style builder and all needed components.

        Create sheet string buffer from _base string template. Copy
        global styles from aird.STYLES and update with class-specific
        styles. Write styles to sheet.
        """
        self.sheet = io.StringIO(
            self._base.format(
                cls=f".{self.sheetclass} " if self.sheetclass else ""
            )
        )
        self.sheet.seek(0, io.SEEK_END)
        self.styles = self._make_styles()

    def _make_styles(self) -> dict[str, dict[str, diagram.CSSdef]]:
        styles = diagram.STYLES["__GLOBAL__"].copy()
        try:
            deep_update_dict(
                styles, diagram.STYLES[self.class_]  # type: ignore[index]
            )
        except KeyError:
            logger.error(
                "No styling defined for diagram class %s", self.class_
            )

        return styles

    def _write_styles_box(
        self,
        elmclass: str,
        pseudo: str,
        styles: dict[str, diagram.CSSdef],
    ) -> None:
        selectors = [
            f"g.Box{elmclass}{pseudo} > {tag}" for tag in ("rect", "use")
        ]
        selector_text = f"g.Box{elmclass}{pseudo} > text"
        if "stroke" in styles:
            self._write_styledict(
                elmclass,
                f"g.Box{elmclass}{pseudo} > line",
                {"stroke": styles["stroke"]},
            )
        self._write_styles_common(elmclass, selectors, selector_text, styles)

    def _write_styles_edge(
        self,
        elmclass: str,
        pseudo: str,
        styles: dict[str, diagram.CSSdef],
    ) -> None:
        selector = f"g.Edge{elmclass}{pseudo} > path"
        selector_text = f"g.Edge{elmclass}{pseudo} > text"
        self._write_styledict(
            elmclass,
            f"g.Edge{elmclass}{pseudo} > rect",
            {"fill": None, "stroke": None},
        )
        self._write_styledict(
            elmclass,
            f"g.Circle{elmclass}{pseudo} > circle",
            {
                "fill": styles.get("stroke", diagram.RGB(0, 0, 0)),
                "stroke": None,
            },
        )

        if not pseudo and elmclass in self._highlight_on_hover:
            self._write_styledict(
                elmclass,
                f"g.Edge{elmclass}:hover > path",
                {
                    "stroke": self._highlight_on_hover[elmclass],
                    "stroke-width": 2,
                },
            )
            self._write_styledict(
                elmclass,
                f"g.Edge{elmclass}:hover > rect",
                {"stroke": self._highlight_on_hover[elmclass]},
            )
        self._write_styles_common(elmclass, selector, selector_text, styles)

    def _write_styles_common(
        self,
        elmclass: str,
        sel_obj: str | list[str],
        sel_text: str,
        allstyles: dict[str, diagram.CSSdef],
    ) -> None:
        for selector, styles in zip(
            [sel_obj, sel_text], _splitstyles(allstyles)
        ):
            if styles:
                self._write_styledict(elmclass, selector, styles)

    def _write_styledict(
        self,
        elmclass: str,
        selector: str | list[str],
        styles: dict[str, diagram.CSSdef],
    ) -> None:
        if isinstance(selector, str):
            selector = f".{self.sheetclass} {selector}"
        else:
            selector = ", ".join(f".{self.sheetclass} {i}" for i in selector)
        self.sheet.write(f"{selector} {{ ")
        self.sheet.write(
            " ".join(
                self._serialize_value(k, v, elmclass)
                for k, v in styles.items()
                if k not in CUSTOM_STYLE_ATTRS
            )
        )
        self.sheet.write(" }\n")

    def _serialize_value(
        self, key: str, value: diagram.CSSdef, class_: str
    ) -> str:
        if key in {"marker-start", "marker-end"}:
            diagram_class = self.class_
            mystyle = diagram.get_style(diagram_class, f"Edge{class_}")
            if "stroke" not in mystyle:
                mystyle["stroke"] = "#f00"

            marker_id = Styling._generate_id(
                str(value),
                [mystyle["stroke"]],
            )
            value = f"url(#{marker_id})"
        elif value is None:
            value = "none"
        elif isinstance(value, diagram.RGB):
            value = str(value)
        elif isinstance(value, cabc.Sequence) and len(value) == 2:
            gradname = f"{class_}{key.capitalize()}Gradient"
            gradcolors = tuple(str(v) for v in value)
            if gradname in self.gradients:
                assert self.gradients[gradname] == gradcolors
            else:
                self.gradients[gradname] = gradcolors
            value = f'url("#{gradname}")'
        elif not isinstance(value, (float, int, str)):
            raise ValueError(f"Invalid stylesheet value: {value}")
        return f"{key}: {value};"


def _splitstyles(
    styles: dict[str, diagram.CSSdef]
) -> tuple[dict[str, diagram.CSSdef], dict[str, diagram.CSSdef]]:
    objstyles: dict[str, diagram.CSSdef] = {}
    textstyles: dict[str, diagram.CSSdef] = {}
    for key, value in styles.items():
        if key.startswith("text_"):
            textstyles[key[len("text_") :]] = value
        else:
            objstyles[key] = value
    return objstyles, textstyles


def deep_update_dict(
    target: cabc.MutableMapping, updates: cabc.Mapping
) -> cabc.MutableMapping:
    """Apply the ``updates`` to the nested ``target`` dict recursively.

    Parameters
    ----------
    target
        The target dict, which will be modified in place.
    updates
        A dict containing the updates to apply. If one of the values is
        the special object ``deep_update_dict.delete``, the
        corresponding key will be deleted from the target.

    Returns
    -------
    dict
        The target dict, after modifying it in place.
    """
    for key, value in list(updates.items()):
        if value is deep_update_dict.delete:  # type: ignore
            del target[key]
        elif isinstance(value, cabc.Mapping):
            target[key] = deep_update_dict(target.get(key, {}), value)
        else:
            target[key] = value
    return target


deep_update_dict.delete = object()  # type: ignore



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Tests for creating and deleting model elements."""
# pylint: disable=missing-function-docstring, redefined-outer-name
import pathlib

import pytest

import capellambse
import capellambse.model as metamodel
import capellambse.model.common as c

# pylint: disable-next=relative-beyond-top-level, unused-import
from .conftest import model as model50  # type: ignore[import]

TEST_ROOT = pathlib.Path(__file__).parent / "data" / "writemodel"
TEST_MODEL = "WriteTestModel.aird"

XPATH_UUID = "//*[@id={!r}]"


@pytest.fixture
def model():
    return capellambse.MelodyModel(TEST_ROOT / TEST_MODEL)


def test_created_elements_can_be_accessed_in_model(
    model: capellambse.MelodyModel,
):
    newobj = model.la.root_component.components.create(name="TestComponent")

    assert newobj is not None
    assert isinstance(newobj, capellambse.model.layers.la.LogicalComponent)
    assert newobj in model.la.root_component.components


def test_created_elements_show_up_in_xml_after_adding_them(
    model: capellambse.MelodyModel,
):
    newobj = model.la.root_component.components.create(name="TestComponent")

    try:
        model._loader[newobj.uuid]
    except KeyError as err:
        raise AssertionError(
            "Cannot find added element via subscripting"
        ) from err

    assert model._loader.xpath(
        XPATH_UUID.format(newobj.uuid)
    ), "Cannot find added element via XPath"


@pytest.mark.parametrize(
    "deletion_target",
    [0, slice(None, 1)],
)
def test_deleted_elements_are_removed(
    model: capellambse.MelodyModel, deletion_target
):
    comps = model.la.root_component.components
    assert len(comps) == 2, "Precondition not met: Bad list length"

    olduuid = comps[0].uuid
    del comps[deletion_target]

    assert len(comps) != 2, "List length did not change"

    with pytest.raises(KeyError):
        model._loader[olduuid]  # pylint: disable=pointless-statement

    assert not model._loader.xpath(
        XPATH_UUID.format(olduuid)
    ), "Element is still present in tree after deleting"


def test_delete_all_deletes_matching_objects(model: capellambse.MelodyModel):
    comps = model.la.root_component.components
    assert len(comps) == 2

    comps.delete_all(name="Delete Me")
    assert len(comps) == 1
    assert comps[0].name == "Keep Me"


def test_create_adds_missing_namespace_to_fragment(
    model: capellambse.MelodyModel,
) -> None:
    assert "Requirements" not in model._element.nsmap, "Precondition failed"
    module = model.by_uuid("85a31dd7-7755-486b-b803-1df8915e2cf9")

    module.requirements.create(name="TestReq")

    assert "Requirements" in model._element.nsmap


def test_adding_a_namespace_preserves_the_capella_version_comment(
    model: capellambse.MelodyModel,
) -> None:
    assert "Requirements" not in model._element.nsmap, "Precondition failed"
    prev_elements = list(model._element.itersiblings(preceding=True))
    assert len(prev_elements) == 1, "No version comment to preserve?"

    model._loader.add_namespace(model._element, "Requirements")

    prev_elements = list(model._element.itersiblings(preceding=True))
    assert len(prev_elements) == 1
    assert model.info.capella_version != "UNKNOWN"


def test_deleting_an_object_purges_references_from_AttrProxyAccessor(
    model: capellambse.MelodyModel, caplog
) -> None:
    part = model.by_uuid("1bd59e23-3d45-4e39-88b4-33a11c56d4e3")
    assert isinstance(part, metamodel.cs.Part)
    assert isinstance(type(part).type, c.AttrProxyAccessor)
    component = model.by_uuid("ea5f09e6-a0ec-46b2-bd3e-b572f9bf99d6")

    component.parent.components.remove(component)

    assert not list(model.find_references(component))
    assert part.type is None
    assert not caplog.records


def test_deleting_an_object_purges_references_from_LinkAccessor(
    model50: capellambse.MelodyModel, caplog
) -> None:
    entity = model50.by_uuid("e37510b9-3166-4f80-a919-dfaac9b696c7")
    assert isinstance(entity, metamodel.oa.Entity)
    assert isinstance(type(entity).activities, c.LinkAccessor)
    activity = model50.by_uuid("f1cb9586-ce85-4862-849c-2eea257f706b")

    activity.parent.activities.remove(activity)

    assert not list(model50.find_references(activity))
    assert activity not in entity.activities
    assert not caplog.records


def test_deleting_an_entire_list_purges_references_to_all_list_members(
    model50: capellambse.MelodyModel, caplog
) -> None:
    component = model50.by_uuid("ff7b8672-84db-4b93-9fea-22a410907fb1")
    assert isinstance(component, metamodel.la.LogicalComponent)
    p1 = model50.by_uuid("db5681e4-4245-4207-a429-e89979f6ac71")
    p2 = model50.by_uuid("7c61d723-3658-47ff-9b0c-7b016ac4cb76")
    current_ports = [i.uuid for i in component.ports]
    assert current_ports == [p1.uuid, p2.uuid], "Unexpected ports"
    assert list(model50.find_references(p1)), "Dead port CP 1?"
    assert list(model50.find_references(p2)), "Dead port CP 2?"

    del component.ports

    assert not list(model50.find_references(p1))
    assert not list(model50.find_references(p2))
    assert not caplog.records


def test_deleting_an_object_purges_references_to_children(
    model50: capellambse.MelodyModel, caplog
) -> None:
    component = model50.by_uuid("a8c46457-a702-41c4-a971-c815c4c5a674")
    port = model50.by_uuid("e0dcf8c2-2283-4456-98a2-146e78ba5f26")
    exchange_id = "d8655737-39ab-4482-a934-ee847c7ff6bd"
    current_refs = [
        (getattr(obj, "uuid", None), attr)
        for obj, attr, _ in model50.find_references(port)
    ]
    assert port in component.ports
    assert len(current_refs) > 1, "Port has no references before test"
    assert (exchange_id, "target") in current_refs

    model50.la.component_package.components.remove(component)

    assert not list(model50.find_references(port))
    assert model50.by_uuid(exchange_id).target is None
    assert not caplog.records
